@@ -12,8 +12,11 @@ class Debugger(var debug: Boolean = false) {
       Console.out
   }
 
+  private val maxOutputLength = 2000
+
   def msg(msg: String) = {
-    out.println("DEBUG: "+msg)
+    if ( msg.length > maxOutputLength ) out.println("DEBUG: "+msg.substring(0,maxOutputLength-3)+"...")
+    else out.println("DEBUG: "+msg)
   }
 
 }
