@@ -28,5 +28,9 @@ object Helper {
     true
   }
 
+  def main(args: Array[String]): Unit = {
+    print(util.Properties.versionNumberString)
+  }
+
 }
 
