@@ -24,11 +24,13 @@ object Protocol {
   val DEBUG         = 103
   val EVAL          = 104
   val SET           = 105
-  val GET           = 106
-  val GET_REFERENCE = 107
-  val DEF           = 108
-  val CLEAR         = 109
-  val INVOKE        = 110
+  val SET_SINGLE    = 106
+  val SET_DOUBLE    = 107
+  val GET           = 108
+  val GET_REFERENCE = 109
+  val DEF           = 110
+  val INVOKE        = 111
+  val SCALAP        = 112
 
   // Result
   val OK = 1000
