@@ -26,7 +26,7 @@ class RClient private (private val in: DataInputStream, private val out: DataOut
     out.flush()
   }
 
-  def eval(snippet: String): Unit = {
+  def eval(snippet: String, evalOnly: Boolean = true): Any = {
     if ( debug ) debugger.msg("Sending EVAL request.")
     out.writeInt(EVAL)
     Helper.writeString(out,snippet)
@@ -35,26 +35,50 @@ class RClient private (private val in: DataInputStream, private val out: DataOut
     val output = Helper.readString(in)
     if ( output != "" ) println(output)
     if ( status != OK ) throw new RuntimeException("Error in R evaluation.")
+    if ( evalOnly ) null else get(".rscala.last.value")._1
   }
 
-  def %~%(snippet: String): Any = {
-    eval(snippet)
-    get(".rscala.last.value")._1
-  }
+  def evalI0(snippet: String) = { eval(snippet,true); getI0(".rscala.last.value") }
+  def evalD0(snippet: String) = { eval(snippet,true); getD0(".rscala.last.value") }
+  def evalB0(snippet: String) = { eval(snippet,true); getB0(".rscala.last.value") }
+  def evalS0(snippet: String) = { eval(snippet,true); getS0(".rscala.last.value") }
+  def evalI1(snippet: String) = { eval(snippet,true); getI1(".rscala.last.value") }
+  def evalD1(snippet: String) = { eval(snippet,true); getD1(".rscala.last.value") }
+  def evalB1(snippet: String) = { eval(snippet,true); getB1(".rscala.last.value") }
+  def evalS1(snippet: String) = { eval(snippet,true); getS1(".rscala.last.value") }
+  def evalI2(snippet: String) = { eval(snippet,true); getI2(".rscala.last.value") }
+  def evalD2(snippet: String) = { eval(snippet,true); getD2(".rscala.last.value") }
+  def evalB2(snippet: String) = { eval(snippet,true); getB2(".rscala.last.value") }
+  def evalS2(snippet: String) = { eval(snippet,true); getS2(".rscala.last.value") }
 
   def selectDynamic(identifier: String): (Any,String) = get(identifier)
 
   def updateDynamic(identifier: String)(value : Any): Unit = set(identifier,value)
 
-  def set(identifier: String, value: Any): Unit = {
+  def set(identifier: String, value: Any, index: String = "", singleBrackets: Boolean = true): Unit = {
     if ( debug ) debugger.msg("Setting: "+identifier)
     val v = value
-    out.writeInt(SET)
+    if ( index == "" ) out.writeInt(SET)
+    else if ( singleBrackets ) {
+      out.writeInt(SET_SINGLE)
+      Helper.writeString(out,index)
+    } else {
+      out.writeInt(SET_DOUBLE)
+      Helper.writeString(out,index)
+    }
     Helper.writeString(out,identifier)
     if ( v == null || v.isInstanceOf[Unit] ) {
       if ( debug ) debugger.msg("... which is null")
       out.writeInt(NULLTYPE)
       out.flush()
+      if ( index != "" ) {
+        val status = in.readInt()
+        if ( status != OK ) {
+          val output = Helper.readString(in)
+          if ( output != "" ) println(output)
+          throw new RuntimeException("Error in R evaluation.")
+        }
+      }
       return
     }
     val c = v.getClass
@@ -168,10 +192,18 @@ class RClient private (private val in: DataInputStream, private val out: DataOut
           out.writeInt(STRING)
           Helper.writeString(out,v.asInstanceOf[String])
         case _ =>
-          throw new RuntimeException("Unsupported array type: "+c.getName)
+          throw new RuntimeException("Unsupported non-array type: "+c.getName)
       }
     }
     out.flush()
+    if ( index != "" ) {
+      val status = in.readInt()
+      if ( status != OK ) {
+        val output = Helper.readString(in)
+        if ( output != "" ) println(output)
+        throw new RuntimeException("Error in R evaluation.")
+      }
+    }
   }
 
   def get(identifier: String): (Any,String) = {
@@ -431,6 +463,8 @@ object RClient {
     cmd.println(raw"q(save='no')")
     cmd.flush()
     val sockets = new ScalaSockets(portsFile.getAbsolutePath,debugger)
+    sockets.out.writeInt(OK)
+    sockets.out.flush()
     apply(sockets.in,sockets.out,debugger)
   }
 
