@@ -2,45 +2,80 @@ package org.ddahl.rscala
 
 import java.net._
 import java.io._
-import scala.tools.nsc.interpreter.IMain
 import scala.Console.{baosOut, baosErr, psOut, psErr, withOut, withErr}
 import scala.annotation.tailrec
 
 import Protocol._
 
-class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugger) {
+class ScalaServer private (repl: InterpreterAdapter, portsFilename: String, debugger: Debugger) {
 
   private val sockets = new ScalaSockets(portsFilename,debugger)
   import sockets.{in, out, socketIn, socketOut}
   private val R = org.ddahl.rscala.RClient(in,out,debugger)
-  repl.bind("R","org.ddahl.rscala.RClient",R)
-  repl.bind("$intp","scala.tools.nsc.interpreter.IMain",repl)
+
+  if ( repl == null ) out.writeInt(ERROR)
+  else {
+    out.writeInt(OK)
+    repl.bind("R","org.ddahl.rscala.RClient",R)
+    if ( repl.interpreter.isInstanceOf[scala.tools.nsc.interpreter.IMain] ) {
+      repl.bind("$intp","scala.tools.nsc.interpreter.IMain",repl.interpreter)
+    }
+  }
+  out.flush
+
   baosOut.reset
   baosErr.reset
 
-  private var functionResult: Any = null
-  private var cacheMap = new scala.collection.mutable.ArrayBuffer[Any]()
+  private var functionResult: (Any, String) = null
+  private var wrapResult: Array[Array[String]] = null
+  private var cacheMap = new scala.collection.mutable.ArrayBuffer[(Any,String)]()
   private val cacheMapExtractor = """\.(\d+)""".r
-  private val functionParameters = new scala.collection.mutable.ArrayBuffer[Any]()
-  private val functionMap = new scala.collection.mutable.HashMap[Any,java.lang.reflect.Method]()
+  private val functionParametersStack = new scala.collection.mutable.ArrayBuffer[Any]()
+  private val functionMap = new scala.collection.mutable.HashMap[String,(Any,java.lang.reflect.Method,Int,String)]()
 
-  private def storeFunction(f: Any): Unit = {
+  private def extractReturnType(signature: String) = {
+    var squareCount = 0
+    var parenCount = 0
+    var i = 0
+    val s = signature.trim
+    while ( ( ( parenCount > 0 ) || ( squareCount > 0 ) || ( ! ( ( s(i) == '=' ) && ( s(i+1) == '>' ) ) ) ) && ( i < s.length ) ) {
+      val c = s(i)
+      i += 1
+      if ( c == '(' ) parenCount += 1
+      if ( c == ')' ) parenCount -= 1
+      if ( c == '[' ) squareCount += 1
+      if ( c == ']' ) squareCount -= 1
+    }
+    if ( i == s.length ) throw new IllegalArgumentException("Unexpected end of signature."+signature)
+    i += 2
+    val r = s.substring(i).trim
+    if ( r.length == 0 ) throw new IllegalArgumentException("Unexpected end of signature."+signature)
+    r
+  }
+
+  private def storeFunction(functionName: String, nArgs: Int): Unit = {
+    val f = repl.valueOfTerm(functionName).get
+    val functionType = repl.typeOfTerm(functionName)
+    val returnType = extractReturnType(functionType)
     val m = f.getClass.getMethods.filter(_.getName == "apply") match {
       case Array() => null
       case Array(method) => method
       case Array(method1,method2) => method1
       case _ => null
     }
-    functionMap(f) = m
+    functionMap(functionName) = (f,m,nArgs,returnType)
   }
 
-  private def callFunction(f: Any): Any = {
+  private def callFunction(functionName: String): Any = {
+    val (f,m,nArgs,resultType) = functionMap(functionName)
+    val functionParameters = functionParametersStack.takeRight(nArgs)
+    functionParametersStack.trimEnd(nArgs)
     if ( debugger.debug ) {
       debugger.msg("Function is: "+f)
       debugger.msg("There are "+functionParameters.length+" parameters.")
       debugger.msg("<<"+functionParameters.mkString(">>\n<<")+">>")
     }
-    functionResult = functionMap(f).invoke(f, functionParameters.map(_.asInstanceOf[AnyRef]): _*)
+    functionResult = (m.invoke(f, functionParameters.map(_.asInstanceOf[AnyRef]): _*),resultType)
   }
 
   private def readString() = Helper.readString(in)
@@ -56,17 +91,12 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
 
   private def setAVM(identifier: String, t: String, v: Any): Unit = {
     if ( debugger.debug ) debugger.msg("Value is "+v)
-    if ( identifier == "." ) functionParameters.append(v)
+    if ( identifier == "." ) functionParametersStack.append(v)
     else repl.bind(identifier,t,v)
   }
 
   private val sep = sys.props("line.separator")
 
-  private def eval(snippet: String): Unit = {
-    repl.interpret(snippet)
-    //snippet.split(sep).foreach(repl.eval)  // Breaking it line-by-line helps hotspot compiler.
-  }
-
   private def doGC(): Unit = {
     if ( debugger.debug ) debugger.msg("Garbage collection")
     val length = in.readInt()
@@ -82,7 +112,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
     try {
       withOut(psOut) {
         withErr(psErr) {
-          eval(snippet)
+          repl.interpret(snippet)
         }
       }
       R.exit()
@@ -101,13 +131,13 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
   }
 
   private def doDef(): Unit = {
-    val signature = readString().trim
+    val args = readString().trim
     val body = readString()
-    if ( debugger.debug ) debugger.msg("Got signature and body.")
+    if ( debugger.debug ) debugger.msg("Got arguments and body.")
     val buffer = new StringBuffer()
     try {
-      val params = if ( signature != "" ) {
-        val y = signature.split(":").map(z => {
+      val params = if ( args != "" ) {
+        val y = args.split(":").map(z => {
           val l = z.lastIndexOf(",")
           if ( l >= 0 ) Array(z.substring(0,l).trim,z.substring(l+1).trim)
           else Array(z.trim)
@@ -121,7 +151,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         }
         (r,s)
       } else (Array[String](),Array[String]())
-      if ( debugger.debug ) debugger.msg("Parsed signature.")
+      if ( debugger.debug ) debugger.msg("Parsed arguments.")
       out.writeInt(OK)
       try {
         val paramNames = params._1
@@ -130,14 +160,14 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         try {
           withOut(psOut) {
             withErr(psErr) {
-              eval(s"($signature) => { $body }")
+              repl.eval(s"($args) => { $body }")
             }
           }
           out.writeInt(OK)
           val functionName = repl.mostRecentVar
           if ( debugger.debug ) debugger.msg("Name of function is: <"+functionName+">")
           try {
-            storeFunction(repl.valueOfTerm(functionName).get)
+            storeFunction(functionName,paramNames.length)
             if ( debugger.debug ) debugger.msg("Stored function.")
             out.writeInt(OK)
             writeString(functionName)
@@ -146,6 +176,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
             if ( debugger.debug ) debugger.msg("There are "+paramNames.length+" parameters.")
             paramNames.foreach(writeString)
             paramTypes.foreach(writeString)
+            writeString(functionMap(functionName)._4)
             if ( debugger.debug ) debugger.msg("Done.")
           } catch {
             case e: Throwable =>
@@ -180,7 +211,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
       val f = repl.valueOfTerm(functionName).get
       withOut(psOut) {
         withErr(psErr) {
-          callFunction(f)
+          callFunction(functionName)
         }
       }
       R.exit()
@@ -198,23 +229,63 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
     out.flush()
   }
 
+  private val RegexpVar = """^\s+(?:final )?\s*(?:override )?\s*var\s+([^(:]+):\s+([^=;]+).*""".r
+  private val RegexpVal = """^\s+(?:final )?\s*(?:override )?\s*val\s+([^(:]+):\s+([^=;]+).*""".r
+  private val RegexpDefNoParen = """^\s+(?:final )?\s*(?:override)?\s*def\s+([a-zA-Z0-9]+)(?:\[.*\]:|:)\s+([^=;]+).*""".r
+  private val RegexpDefWithParen = """^\s+(?:final )?\s*(?:override)?\s*def\s+([a-zA-Z0-9]+)(?:\[.*\]+)?\((.*)\):\s+([^=]+).*""".r
+  private val RegexpConstructor = """^\s+def\s+this\((.*)\).*""".r
+  private val RegexpThrows = """^\s*throws\s+.*""".r
+  private val rscalaReference = "rscalaReference"
+
+  private def parseArgNames(args: String): Array[String] = {
+    if ( args.contains(")(") ) null
+    else {
+      val x = args.split(":")
+      val y = if ( x.length > 1 ) x.dropRight(1) else x
+      y.map(z => {
+        val i = z.lastIndexOf(",")
+        if ( i >= 0 ) z.substring(i+1).trim else z.trim
+      })
+    }
+  }
+
+  private def doScalap(): Unit = {
+    if ( debugger.debug ) debugger.msg("Doing scalap...")
+    val itemName = readString()
+    if ( debugger.debug ) debugger.msg("... on "+itemName)
+    val classpath = sys.props("sun.boot.class.path") + sys.props("path.separator") + sys.props("rscala.classpath")
+    if ( debugger.debug ) debugger.msg("... with classpath "+classpath)
+    import java.io.{ByteArrayOutputStream, PrintStream}
+    val baos = new ByteArrayOutputStream()
+    val ps = new PrintStream(baos,true)
+    scala.Console.withOut(ps) {
+      scala.tools.scalap.Main.main(Array("-cp",classpath,itemName))
+    }
+    writeString(baos.toString)
+    out.flush()
+    baos.close()
+    ps.close()
+  }
+
   private def doSet(): Unit = {
     val identifier = readString()
     val buffer = new StringBuffer()
     in.readInt() match {
       case NULLTYPE =>
         if ( debugger.debug ) debugger.msg("Setting null.")
-        if ( identifier == "." ) functionParameters.append(null)
+        if ( identifier == "." ) functionParametersStack.append(null)
         else repl.bind(identifier,"Any",null)
       case REFERENCE =>
         if ( debugger.debug ) debugger.msg("Setting reference.")
         val originalIdentifier = readString()
         try {
           if ( identifier == "." ) originalIdentifier match {
+            case "null" =>
+              functionParametersStack.append(null)
             case cacheMapExtractor(i) =>
-              functionParameters.append(cacheMap(i.toInt))
+              functionParametersStack.append(cacheMap(i.toInt)._1)
             case _ =>
-              functionParameters.append(repl.valueOfTerm(originalIdentifier).get)
+              functionParametersStack.append(repl.valueOfTerm(originalIdentifier).get)
           } else {
             val r = identifier.split(":") match {
               case Array(n)   => (n.trim,"")
@@ -222,13 +293,13 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
             }
             originalIdentifier match {
               case cacheMapExtractor(i) =>
-                val t = if ( r._2 == "" ) "Any" else r._2
-                repl.bind(r._1,t,cacheMap(i.toInt))
+                val (originalValue,originalType) = cacheMap(i.toInt)
+                repl.bind(r._1,originalType,originalValue)
               case _ =>
                 val vt = if ( r._2 == "" ) r._1 else r._1 + ": " + r._2
                 withOut(psOut) {
                   withErr(psErr) {
-                    eval(s"val ${vt} = ${originalIdentifier}")
+                    repl.eval(s"val ${vt} = ${originalIdentifier}")
                   }
                 }
             }
@@ -286,18 +357,20 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
   private def doGet(): Unit = {
     val identifier = readString()
     if ( debugger.debug ) debugger.msg("Trying to get value of: "+identifier)
-    val option = try {
+    val optionWithType = try {
       identifier match {
         case "." => 
-          repl.valueOfTerm(repl.mostRecentVar)
+          val mrv = repl.mostRecentVar
+          (repl.valueOfTerm(mrv),repl.typeOfTerm(mrv))
         case "?" => 
-          val s = Some(functionResult)
-          functionResult = null
-          s
+          (Some(functionResult._1),functionResult._2)
         case cacheMapExtractor(i) =>
-          Some(cacheMap(i.toInt))
+          val tuple = cacheMap(i.toInt)
+          (Some(tuple._1),tuple._2)
+        case "null" =>
+          (Some(null),"Null")
         case _ =>
-          repl.valueOfTerm(identifier)
+          (repl.valueOfTerm(identifier),repl.typeOfTerm(identifier))
       }
     } catch {
       case e: Throwable =>
@@ -311,134 +384,153 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         out.flush()
         return
     }
-    if ( option.isEmpty ) {
+    if ( debugger.debug ) debugger.msg("Getting: "+identifier)
+    if ( optionWithType._1.isEmpty ) {
+      if ( debugger.debug ) debugger.msg("... which does not exist.")
       out.writeInt(UNDEFINED_IDENTIFIER)
+      val buffer = new StringBuffer()
+      writeAll(buffer)
       out.flush()
       return
     }
-    val v = option.get
-    if ( debugger.debug ) debugger.msg("Getting: "+identifier)
+    val v = optionWithType._1.get
     if ( v == null || v.isInstanceOf[Unit] ) {
       if ( debugger.debug ) debugger.msg("... which is null")
       out.writeInt(NULLTYPE)
-      out.flush()
+      out.flush
       return
     }
-    val c = v.getClass
-    if ( debugger.debug ) debugger.msg("... whose class is: "+c)
+    val t = if ( optionWithType._2 == "Any" ) {
+      val c = v.getClass
+      if ( debugger.debug ) debugger.msg("Trying to infer type of "+c)
+      if ( c.isArray ) {
+        c.getName match {
+          case "[I" => "Array[Int]"
+          case "[D" => "Array[Double]"
+          case "[Z" => "Array[Boolean]"
+          case "[Ljava.lang.String;" => "Array[String]"
+          case "[[I" => "Array[Array[Int]]"
+          case "[[D" => "Array[Array[Double]]"
+          case "[[Z" => "Array[Array[Boolean]]"
+          case "[[Ljava.lang.String;" => "Array[Array[String]]"
+          case _ => "Any"
+        }
+      } else {
+        c.getName match {
+          case "java.lang.Integer" => "Int"
+          case "java.lang.Double" => "Double"
+          case "java.lang.Boolean" => "Boolean"
+          case "java.lang.String" => "String"
+          case _ => "Any"
+        }
+      }
+    } else optionWithType._2
+    if ( debugger.debug ) debugger.msg("... whose type is: "+t)
     if ( debugger.debug ) debugger.msg("... and whose value is: "+v)
-    if ( c.isArray ) {
-      c.getName match {
-        case "[I" =>
-          val vv = v.asInstanceOf[Array[Int]]
-          out.writeInt(VECTOR)
+    t match {
+      case "Array[Int]" =>
+        val vv = v.asInstanceOf[Array[Int]]
+        out.writeInt(VECTOR)
+        out.writeInt(vv.length)
+        out.writeInt(INTEGER)
+        for ( i <- 0 until vv.length ) out.writeInt(vv(i))
+      case "Array[Double]" =>
+        val vv = v.asInstanceOf[Array[Double]]
+        out.writeInt(VECTOR)
+        out.writeInt(vv.length)
+        out.writeInt(DOUBLE)
+        for ( i <- 0 until vv.length ) out.writeDouble(vv(i))
+      case "Array[Boolean]" =>
+        val vv = v.asInstanceOf[Array[Boolean]]
+        out.writeInt(VECTOR)
+        out.writeInt(vv.length)
+        out.writeInt(BOOLEAN)
+        for ( i <- 0 until vv.length ) out.writeInt(if ( vv(i) ) 1 else 0)
+      case "Array[String]" =>
+        val vv = v.asInstanceOf[Array[String]]
+        out.writeInt(VECTOR)
+        out.writeInt(vv.length)
+        out.writeInt(STRING)
+        for ( i <- 0 until vv.length ) writeString(vv(i))
+      case "Array[Array[Int]]" =>
+        val vv = v.asInstanceOf[Array[Array[Int]]]
+        if ( Helper.isMatrix(vv) ) {
+          out.writeInt(MATRIX)
           out.writeInt(vv.length)
+          if ( vv.length > 0 ) out.writeInt(vv(0).length)
+          else out.writeInt(0)
           out.writeInt(INTEGER)
-          for ( i <- 0 until vv.length ) out.writeInt(vv(i))
-        case "[D" =>
-          val vv = v.asInstanceOf[Array[Double]]
-          out.writeInt(VECTOR)
-          out.writeInt(vv.length)
-          out.writeInt(DOUBLE)
-          for ( i <- 0 until vv.length ) out.writeDouble(vv(i))
-        case "[Z" =>
-          val vv = v.asInstanceOf[Array[Boolean]]
-          out.writeInt(VECTOR)
-          out.writeInt(vv.length)
-          out.writeInt(BOOLEAN)
-          for ( i <- 0 until vv.length ) out.writeInt(if ( vv(i) ) 1 else 0)
-        case "[Ljava.lang.String;" =>
-          val vv = v.asInstanceOf[Array[String]]
-          out.writeInt(VECTOR)
-          out.writeInt(vv.length)
-          out.writeInt(STRING)
-          for ( i <- 0 until vv.length ) writeString(vv(i))
-        case "[[I" =>
-          val vv = v.asInstanceOf[Array[Array[Int]]]
-          if ( Helper.isMatrix(vv) ) {
-            out.writeInt(MATRIX)
-            out.writeInt(vv.length)
-            if ( vv.length > 0 ) out.writeInt(vv(0).length)
-            else out.writeInt(0)
-            out.writeInt(INTEGER)
-            for ( i <- 0 until vv.length ) {
-              val vvv = vv(i)
-              for ( j <- 0 until vvv.length ) {
-                out.writeInt(vv(i)(j))
-              }
+          for ( i <- 0 until vv.length ) {
+            val vvv = vv(i)
+            for ( j <- 0 until vvv.length ) {
+              out.writeInt(vv(i)(j))
             }
           }
-        case "[[D" =>
-          val vv = v.asInstanceOf[Array[Array[Double]]]
-          if ( Helper.isMatrix(vv) ) {
-            out.writeInt(MATRIX)
-            out.writeInt(vv.length)
-            if ( vv.length > 0 ) out.writeInt(vv(0).length)
-            else out.writeInt(0)
-            out.writeInt(DOUBLE)
-            for ( i <- 0 until vv.length ) {
-              val vvv = vv(i)
-              for ( j <- 0 until vvv.length ) {
-                out.writeDouble(vvv(j))
-              }
-            }
-          } else out.writeInt(UNSUPPORTED_STRUCTURE)
-        case "[[Z" =>
-          val vv = v.asInstanceOf[Array[Array[Boolean]]]
-          if ( Helper.isMatrix(vv) ) {
-            out.writeInt(MATRIX)
-            out.writeInt(vv.length)
-            if ( vv.length > 0 ) out.writeInt(vv(0).length)
-            else out.writeInt(0)
-            out.writeInt(BOOLEAN)
-            for ( i <- 0 until vv.length ) {
-              val vvv = vv(i)
-              for ( j <- 0 until vv(i).length ) {
-                out.writeInt(if ( vvv(j) ) 1 else 0)
-              }
-            }
-          } else out.writeInt(UNSUPPORTED_STRUCTURE)
-        case "[[Ljava.lang.String;" =>
-          val vv = v.asInstanceOf[Array[Array[String]]]
-          if ( Helper.isMatrix(vv) ) {
-            out.writeInt(MATRIX)
-            out.writeInt(vv.length)
-            if ( vv.length > 0 ) out.writeInt(vv(0).length)
-            else out.writeInt(0)
-            out.writeInt(STRING)
-            for ( i <- 0 until vv.length ) {
-              val vvv = vv(i)
-              for ( j <- 0 until vv(i).length ) {
-                writeString(vvv(j))
-              }
-            }
-          } else out.writeInt(UNSUPPORTED_STRUCTURE)
-        case _ =>
-          if ( debugger.debug ) debugger.msg("Bowing out: "+identifier)
-          out.writeInt(UNSUPPORTED_STRUCTURE)
-      }
-    } else {
-      c.getName match {
-        case "java.lang.Integer" =>
-          out.writeInt(ATOMIC)
-          out.writeInt(INTEGER)
-          out.writeInt(v.asInstanceOf[Int])
-        case "java.lang.Double" =>
-          out.writeInt(ATOMIC)
+        }
+      case "Array[Array[Double]]" =>
+        val vv = v.asInstanceOf[Array[Array[Double]]]
+        if ( Helper.isMatrix(vv) ) {
+          out.writeInt(MATRIX)
+          out.writeInt(vv.length)
+          if ( vv.length > 0 ) out.writeInt(vv(0).length)
+          else out.writeInt(0)
           out.writeInt(DOUBLE)
-          out.writeDouble(v.asInstanceOf[Double])
-        case "java.lang.Boolean" =>
-          out.writeInt(ATOMIC)
+          for ( i <- 0 until vv.length ) {
+            val vvv = vv(i)
+            for ( j <- 0 until vvv.length ) {
+              out.writeDouble(vvv(j))
+            }
+          }
+        } else out.writeInt(UNSUPPORTED_STRUCTURE)
+      case "Array[Array[Boolean]]" =>
+        val vv = v.asInstanceOf[Array[Array[Boolean]]]
+        if ( Helper.isMatrix(vv) ) {
+          out.writeInt(MATRIX)
+          out.writeInt(vv.length)
+          if ( vv.length > 0 ) out.writeInt(vv(0).length)
+          else out.writeInt(0)
           out.writeInt(BOOLEAN)
-          out.writeInt(if (v.asInstanceOf[Boolean]) 1 else 0)
-        case "java.lang.String" =>
-          out.writeInt(ATOMIC)
+          for ( i <- 0 until vv.length ) {
+            val vvv = vv(i)
+            for ( j <- 0 until vv(i).length ) {
+              out.writeInt(if ( vvv(j) ) 1 else 0)
+            }
+          }
+        } else out.writeInt(UNSUPPORTED_STRUCTURE)
+      case "Array[Array[String]]" =>
+        val vv = v.asInstanceOf[Array[Array[String]]]
+        if ( Helper.isMatrix(vv) ) {
+          out.writeInt(MATRIX)
+          out.writeInt(vv.length)
+          if ( vv.length > 0 ) out.writeInt(vv(0).length)
+          else out.writeInt(0)
           out.writeInt(STRING)
-          writeString(v.asInstanceOf[String])
-        case _ =>
-          if ( debugger.debug ) debugger.msg("Bowing out: "+identifier)
-          out.writeInt(UNSUPPORTED_STRUCTURE)
-      }
+          for ( i <- 0 until vv.length ) {
+            val vvv = vv(i)
+            for ( j <- 0 until vv(i).length ) {
+              writeString(vvv(j))
+            }
+          }
+        } else out.writeInt(UNSUPPORTED_STRUCTURE)
+      case "Int" =>
+        out.writeInt(ATOMIC)
+        out.writeInt(INTEGER)
+        out.writeInt(v.asInstanceOf[Int])
+      case "Double" =>
+        out.writeInt(ATOMIC)
+        out.writeInt(DOUBLE)
+        out.writeDouble(v.asInstanceOf[Double])
+      case "Boolean" =>
+        out.writeInt(ATOMIC)
+        out.writeInt(BOOLEAN)
+        out.writeInt(if (v.asInstanceOf[Boolean]) 1 else 0)
+      case "String" =>
+        out.writeInt(ATOMIC)
+        out.writeInt(STRING)
+        writeString(v.asInstanceOf[String])
+      case _ =>
+        if ( debugger.debug ) debugger.msg("Bowing out: "+identifier)
+        out.writeInt(UNSUPPORTED_STRUCTURE)
     }
     out.flush()
   }
@@ -451,13 +543,20 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         cacheMap.append(functionResult)
         out.writeInt(OK)
         writeString("." + (cacheMap.length-1))
-        writeString(functionResult.getClass.getName)
+        writeString(functionResult._2)
       case cacheMapExtractor(i) =>
         out.writeInt(OK)
         writeString(identifier)
         val r = cacheMap(i.toInt)
         writeString(r.getClass.getName)
       case _ =>
+        if ( identifier == "null" ) {
+          out.writeInt(OK)
+          writeString("null")
+          writeString("Null")
+          out.flush()
+          return
+        }
         val id = if ( identifier == "." ) repl.mostRecentVar else identifier
         val option = try {
           repl.valueOfTerm(id)
@@ -477,7 +576,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         else {
           out.writeInt(OK)
           writeString(id)
-          writeString(repl.typeOfTerm(id).toString)
+          writeString(repl.typeOfTerm(id))
         }
     }
     out.flush()
@@ -485,6 +584,14 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
 
   @tailrec
   final def run(): Unit = {
+    if ( repl == null ) {
+      if ( debugger.debug ) debugger.msg("Exiting")
+      in.close()
+      out.close()
+      socketIn.close()
+      socketOut.close()
+      return
+    }
     if ( debugger.debug ) debugger.msg("Top of the loop waiting for a command.")
     val request = try {
       in.readInt()
@@ -520,10 +627,10 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
         doGetReference()
       case DEF =>
         doDef()
-      case CLEAR =>
-        functionParameters.clear
       case INVOKE =>
         doInvoke()
+      case SCALAP =>
+        doScalap()
     }
     run()
   }
@@ -532,7 +639,7 @@ class ScalaServer private (repl: IMain, portsFilename: String, debugger: Debugge
 
 object ScalaServer {
 
-  def apply(repl: IMain, portsFilename: String, debug: Boolean = false): ScalaServer = new ScalaServer(repl, portsFilename, new Debugger(debug))
+  def apply(repl: InterpreterAdapter, portsFilename: String, debug: Boolean = false): ScalaServer = new ScalaServer(repl, portsFilename, new Debugger(debug))
 
 }
 
